| C prototype: void FastDrawHLine_R(register void* plane asm("a0"), register short x1 asm("d0"), register short x2 asm("d1"), register short y asm("d2"),short mode) __attribute__((__stkparm__));
| Valid values for mode are: A_REVERSE, A_NORMAL, A_XOR, A_REPLACE, A_OR.
|
| This routine draws a horizontal line from (x1;y) to (x2;y).
| This is the version which receives most parameters in the registers, not
| on the stack. plane goes in a0, x1 in d0, x2 in d1, y in d2, mode on the
| stack.
|
| Many French people often criticize ExtGraph for its slowness. Of course,
| ExtGraph is slower than other libraries, if the equivalent routines in these
| other libraries are written with completely unrolled loops. The comparison
| can't be done in an impartial way, but some of those people do it anyway...
| And they use this kind of arguments to advertise for their own projects (but
| indeed, they don't mention that the comparison is pointless; they don't
| mention either that their projects have serious drawbacks ExtGraph doesn't
| have)... And the other libraries are written in assembly, whereas ExtGraph
| is (was) nearly completely written in C.
| Needless to say that these people are not willing to improve it, even if
| they have the skills required to do it (some are more skilled than I am)...
|
| As a result, we are really happy to see that good programmers are not among
| the people mentioned above.
| This routine was contributed to ExtGraph by Julien Richard-Foy
| ('jackiechan'). It is much faster than the previous routine, especially this
| version which takes its parameters into the registers.


.data
.even
__Beginning_FDHL_R__:
.word 0xFFFF,0x7FFF,0x3FFF,0x1FFF,0x0FFF,0x07FF,0x03FF,0x01FF,0x00FF,0x007F,0x003F,0x001F,0x000F,0x0007,0x0003,0x0001

__End_FDHL_R__:
.word 0x8000,0xC000,0xE000,0xF000,0xF800,0xFC00,0xFE00,0xFF00,0xFF80,0xFFC0,0xFFE0,0xFFF0,0xFFF8,0xFFFC,0xFFFE,0xFFFF

.globl FastDrawHLine_R
FastDrawHLine_R:
    move.l   %d4,-(%sp)                      | d4 mustn't be destroyed.

    cmp.w    %d0,%d1
    bge.s    __Registers_well_ordered_FDHL_R__

    exg      %d0,%d1

__Registers_well_ordered_FDHL_R__:
    move.w   %d2,%d4
    lsl.w    #4,%d2
    sub.w    %d4,%d2
    move.w   %d0,%d4
    lsr.w    #4,%d4
    add.w    %d4,%d2
    add.w    %d2,%d2
    adda.w   %d2,%a0

| d4 = 8 * (x1/16 + x1/16) + 16. We add 1 before shifting instead of adding 16 
| after shifting (gain: 4 clocks and 2 bytes).
    addq.w   #1,%d4                          | d4 = 8 * (x1/16 + x1/16) + 16.
    lsl.w    #4,%d4

    move.w   %d1,%d2                         | x2 is stored in d2.
    andi.w   #0xF,%d0

    add.w    %d0,%d0
    move.w   __Beginning_FDHL_R__(%pc,%d0.w),%d0 | d0 = mask of first pixels.
    andi.w   #0xF,%d1

    add.w    %d1,%d1
    move.w   __End_FDHL_R__(%pc,%d1.w),%d1   | d1 = mask of last pixels.
    cmp.w    %d4,%d2                         | Line less than 16 pixels long ?
    blt.s    __Less_than_16_pixels_FDHL_R__
    sub.w    %d4,%d2                         | d2 = x2 - x.
    moveq.l  #32,%d4
    tst.w    4+4(%sp)
    beq.s    __Reverse_FDHL_R__
    cmpi.w   #2,4+4(%sp)
    beq.s    __Xor_FDHL_R__

__Normal_FDHL_R__:
    or.w     %d0,(%a0)+
    moveq    #-1,%d0
    sub.w    %d4,%d2
    blt.s    __Normal_16_pixels_FDHL_R__
__Normal_32_pixels_FDHL_R__:
    move.l   %d0,(%a0)+
    sub.w    %d4,%d2
    bge.s    __Normal_32_pixels_FDHL_R__
__Normal_16_pixels_FDHL_R__:
    cmpi.w   #-16,%d2
    blt.s    __Normal_lessthan16_pixels_FDHL_R__
    move.w   %d0,(%a0)+
__Normal_lessthan16_pixels_FDHL_R__:
    or.w     %d1,(%a0)
    move.l   (%sp)+,%d4
    rts

__Reverse_FDHL_R__:
    not.w    %d0
    and.w    %d0,(%a0)+
    moveq    #0,%d0
    sub.w    %d4,%d2
    blt.s    __Reverse_16_pixels_FDHL_R__
__Reverse_32_pixels_FDHL_R__:
    move.l   %d0,(%a0)+
    sub.w    %d4,%d2
    bge.s    __Reverse_32_pixels_FDHL_R__
__Reverse_16_pixels_FDHL_R__:
    cmpi.w   #-16,%d2
    blt.s    __Reverse_lessthan16_pixels_FDHL_R__
    move.w   %d0,(%a0)+
__Reverse_lessthan16_pixels_FDHL_R__:
    not.w    %d1
    and.w    %d1,(%a0)
    move.l   (%sp)+,%d4
    rts

__Xor_FDHL_R__:
    eor.w    %d0,(%a0)+
    sub.w    %d4,%d2
    blt.s    __Xor_16_pixels_FDHL_R__
__Xor_32_pixels_FDHL_R__:
    not.l    (%a0)+
    sub.w    %d4,%d2
    bge.s    __Xor_32_pixels_FDHL_R__
__Xor_16_pixels_FDHL_R__:
    cmpi.w   #-16,%d2
    blt.s    __Xor_lessthan16_pixels_FDHL_R__
    not.w    (%a0)+
__Xor_lessthan16_pixels_FDHL_R__:
    eor.w    %d1,(%a0)
    move.l   (%sp)+,%d4
    rts

__Less_than_16_pixels_FDHL_R__:
    move.l   (%sp)+,%d4
    and.w    %d0,%d1
    tst.w    4(%sp)
    beq.s    __Reverse_l16_FDHL_R__
    cmpi.w   #2,4(%sp)
    beq.s    __Xor_l16_FDHL_R__
    or.w     %d1,(%a0)
    rts
__Reverse_l16_FDHL_R__:
    not.w    %d1
    and.w    %d1,(%a0)
    rts
__Xor_l16_FDHL_R__:
    eor.w    %d1,(%a0)
    rts
